/*
 * We build up a tree of TclObj_Signature for each class. A node of
 * the tree is one of the member function, either native or inherited
 * by one of its base classes. Only one node is created for each
 * name.
 * Overloaded functions are then resolved in the TclObj_Argtree. If
 * the function's parameter matches the identifier in the Argtree
 * node, go down. Otherwise, follow the tree left or right. Once
 * the parameter list is exhausted and down is NULL, we can find
 * the provider of this function.
 * The nodes must be inserted in a breadth-first traversal of the
 * class hierarchy, so that functions with the same parameter list
 * are properly dominated.
 */

#include <tcl.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "tclobj.h"
#include "signature.h"
#include "register.h"
#include "compat.h"
#include "fptools.h"

/*
 * cvs version
 */
char * tclobj_signature_id = "$Id: signature.cc,v 1.2 1996/11/28 15:12:12 fp Exp $";

/*
 * Take a class name and return its identifier
 */

unsigned long
TclObj_GetClassIdent (Tcl_Interp *interp, const char *arg)
{
  const TclObj_Info *theinfo;
  unsigned long result;

  if (arg[0] == '.')
    result = TCL_OBJECT_ARGTYPE_VARARG;
  else if (_FP_stricmp ((char *) arg, "string") == 0)
    result = TCL_OBJECT_ARGTYPE_STRING;
  else if (_FP_stricmp ((char *) arg, "bool") == 0)
    result = TCL_OBJECT_ARGTYPE_BOOL;
  else if (_FP_stricmp ((char *) arg, "double") == 0)
    result = TCL_OBJECT_ARGTYPE_DOUBLE;
  else if (_FP_stricmp ((char *) arg, "int") == 0)
    result = TCL_OBJECT_ARGTYPE_INT;
  else if (_FP_stricmp ((char *) arg, "uint") == 0)
    result = TCL_OBJECT_ARGTYPE_UINT;
  else if (_FP_stricmp ((char *) arg, "long") == 0)
    result = TCL_OBJECT_ARGTYPE_LONG;
  else if (_FP_stricmp ((char *) arg, "ulong") == 0)
    result = TCL_OBJECT_ARGTYPE_ULONG;
  else if (_FP_stricmp ((char *) arg, "ptr") == 0)
    result = TCL_OBJECT_ARGTYPE_PTR;
  else /* any other object */ {
    if ((theinfo = TclObj_GetClassInfo (interp, arg)) == NULL) {
      /*
       * not yet available, try to find it
       */
      if ((theinfo = TclObj_Require (interp, arg)) == NULL)
	return 0;
    }
    /*
     * got info
     */
    result = theinfo->ident;
  }
  return result;
}

static TclObj_Argtree *
TclObj_AddArgtree (Tcl_Interp *interp,
		   TclObj_Argtree *here,
		   const TclObj_Info *provider,
		   const char *arglist,
		   const TclObj_Comtab *entry,
		   int *result)
{
  TclObj_Argtree *node=here, *iter;
  const char *ptr = arglist;
  char tmpstr[64], *tptr;
  unsigned long ident;


  while (isspace (*ptr)) ptr++;

  /*
   * if no symbols are left in the arglist, make a node.
   */

  if (*ptr == '\0') {
    if (here == NULL) {
      if ((node = new TclObj_Argtree) == NULL) {
	Tcl_AppendResult (interp, "out of memory in TclObj_AddArgtree",
			  NULL);
	*result = TCL_ERROR;
	return NULL;
      }
      node->ident    = 0; /* no args left */
      node->provider = provider;
      node->entry    = entry;
      node->next     = NULL;
      node->down     = NULL;
    }
    else {
      /*
       * another function with the same parameters up to this point.
       * see if there is another function that also doesn't have any
       * more parameters.
       */
      for (iter=here; iter->next&&iter->ident!=0; iter=iter->next);

      if (iter->ident == 0) {
	/*
	 * another function with the same parameter list.
	 * if the provider is the same, ignore me (multiple
	 * inheritance!), otherwise add me as next
	 */
	if (provider->ident != iter->provider->ident) {
	  if ((iter = new TclObj_Argtree) == NULL) {
	    Tcl_AppendResult (interp, "out of memory in TclObj_AddArgtree",
			      NULL);
	    *result = TCL_ERROR;
	    return here;
	  }
	  iter->ident    = 0;
	  iter->provider = provider;
	  iter->entry    = entry;
	  iter->down     = NULL;

	  iter->next = here->next;
	  here->next = iter;
	}
      }
      else {
	/*
	 * there is another function with more arguments. hook me up
	 */
	if ((iter->next = new TclObj_Argtree) == NULL) {
	  Tcl_AppendResult (interp, "out of memory in TclObj_AddArgtree",
			    NULL);
	  *result = TCL_ERROR;
	  return here;
	}
	iter->next->ident    = 0;
	iter->next->provider = provider;
	iter->next->entry    = entry;
	iter->next->next     = NULL;
	iter->next->down     = NULL;
      }
      node = here;
    }
    return node;
  }

  /*
   * not yet at the end of the parameter list. keep going.
   */

  tptr = tmpstr;
  while (*ptr == '.' || *ptr == '*')   *tptr++ = *ptr++;
  while (*ptr == '_' || isalpha(*ptr)) *tptr++ = *ptr++;
  while (*ptr == '_' || isalnum(*ptr)) *tptr++ = *ptr++;
  *tptr = '\0';
  while (isspace (*ptr)) ptr++;
  if (*ptr==',') ptr++;
  while (isspace (*ptr)) ptr++;

  if (tmpstr[0] == '\0') {
    Tcl_AppendResult (interp, "error parsing argument list", NULL);
    *result = TCL_ERROR;
    return NULL;
  }

  if ((ident = TclObj_GetClassIdent (interp, tmpstr)) == 0) {
    *result = TCL_ERROR;
    return NULL;
  }

  if (here == NULL) {
    if ((node = new TclObj_Argtree) == NULL) {
      Tcl_AppendResult (interp, "out of memory in TclObj_AddArgtree",
			NULL);
      *result = TCL_ERROR;
      return NULL;
    }

    node->ident    = ident;
    node->provider = NULL;
    node->entry    = NULL;
    node->next     = NULL;
    node->down     = TclObj_AddArgtree (interp, NULL, provider,
					ptr, entry, result);
  }
  else {
    /*
     * another function with the same parameters up to this point.
     * check if this parameter is the same, too.
     */
    for (iter=here; iter->next&&iter->ident!=ident; iter=iter->next);
    
    if (iter->ident == ident) {
      /*
       * same parameter as an already available func
       */
      iter->down = TclObj_AddArgtree (interp, iter->down, provider,
				      ptr, entry, result);
    }
    else {
      /*
       * the other functions all have a different argument here
       */
      if ((iter->next = new TclObj_Argtree) == NULL) {
	Tcl_AppendResult (interp, "out of memory in TclObj_AddArgtree",
			  NULL);
	*result = TCL_ERROR;
	return here;
      }
      iter->next->ident    = ident;
      iter->next->provider = NULL;
      iter->next->entry    = NULL;
      iter->next->next     = NULL;
      iter->next->down     = TclObj_AddArgtree (interp, NULL, provider,
						ptr, entry, result);
    }
    node = here;
  }

  return node;
}

static TclObj_Signature *
TclObj_AddSignature (Tcl_Interp *interp,
		     TclObj_Signature *here,
		     const TclObj_Info *provider,
		     const char *name,
		     const char *arglist,
		     const TclObj_Comtab *entry,
		     int *result)
{
  TclObj_Signature *node;
  int cres;

  if (here == NULL) {
    if ((node = new TclObj_Signature) == NULL) {
      Tcl_AppendResult (interp, "out of memory in TclObj_AddSignature",
			NULL);
      *result = TCL_ERROR;
      return NULL;
    }
    if ((node->name = new char[strlen(name)+1]) == NULL) {
      Tcl_AppendResult (interp, "out of memory in TclObj_AddSignature",
			NULL);
      *result = TCL_ERROR;
      delete node;
      return NULL;
    }

    strcpy (node->name, name);
    node->left = node->right = NULL;

    node->argtree = TclObj_AddArgtree (interp, NULL, provider,
				       arglist, entry, result);

    if (*result == TCL_ERROR) {
      Tcl_AppendResult (interp, " while processing ",
			(char *) provider->name, "::",
			(char *) name, "\n", NULL);
    }
  }
  else {
    cres = _FP_stricmp ((char *) name, here->name);
    node = here;

    if (cres < 0) {
      here->left  = TclObj_AddSignature (interp, here->left, provider,
					 name, arglist, entry, result);
    }
    else if (cres > 0) {
      here->right = TclObj_AddSignature (interp, here->right, provider,
					 name, arglist, entry, result);
    }
    else /* if (cres == 0) */ {
      here->argtree = TclObj_AddArgtree (interp, here->argtree, provider,
					 arglist, entry, result);
      if (*result == TCL_ERROR) {
	Tcl_AppendResult (interp, " while processing ",
			  (char *) provider->name, "::",
			  (char *) name, "\n", NULL);
      }
    }
  }
  return node;
}

/*
 * browse class information and its base classes and build up the
 * signature tree
 */

TclObj_Signature *
TclObj_BuildSigtree (Tcl_Interp *interp,
		     TclObj_Signature *data,
		     const TclObj_Info *theinfo,
		     int *errcode, int level)
{
  TclObj_Signature *result=data;
  const TclObj_Comtab *iter;

  /*
   * Browse new table and command table. Do not add constructors of
   * base classes.
   */

  if (theinfo->comtab) {
    for (iter=theinfo->comtab; iter->number; iter++) {
      if (level==0 || _FP_stricmp (iter->name, "new") != 0)
	result = TclObj_AddSignature (interp, result, theinfo,
				      iter->name, iter->args,
				      iter, errcode);
    }
  }

  return result;
}

/*
 * ----------------------------------------------------------------------
 * Traverse the tree of available signatures checking against
 * a particular list of arguments. We need to know the provider
 * and the command number, so we just return the complete
 * TclObj_Argtree structure.
 * If scopeident is non-NULL, it is the identifier of a class
 * specified with a scope-resolution operator
 * ----------------------------------------------------------------------
 */

static const TclObj_Argtree *
TclObj_SearchArgtree (Tcl_Interp *interp,
		      const TclObj_Argtree *here,
		      unsigned long scopeident,
		      int argc, char *argv[],
		      int *result)
{
  const TclObj_Argtree *iter;
  Tcl_CmdInfo cmdinfo;
  int itemp, found;
  double dtemp;
  void *vtemp;

  if (here == NULL)
    return NULL;

  if (argc == 0) {
    /*
     * no arguments left. see if there is a node in the Argtree that
     * doesn't want more parameters, too. This can be an empty list
     * or TCL_OBJECT_ARGTYPE_VARARG
     */
    for (iter=here; iter; iter=iter->next) {
      if (iter->ident == 0) {
	if (scopeident==0 || scopeident == iter->provider->ident)
	  return iter;
      }
      else if (iter->ident == TCL_OBJECT_ARGTYPE_VARARG) {
	if (scopeident==0 || scopeident==iter->down->provider->ident)
	  return iter->down;
      }
    }

    /*
     * if we got here, we didn't find anything
     */

    return NULL;
  }

  /*
   * determine type of current argument
   */

  for (iter=here, found=0; iter; iter=iter->next) {
    switch (iter->ident) {
    case 0:
      break;

    case TCL_OBJECT_ARGTYPE_VARARG:
      if (iter->down && iter->down->ident==0 && iter->down->provider &&
	  (scopeident==0 || scopeident==iter->down->provider->ident)) {
	return iter->down;
      }
      break;

    case TCL_OBJECT_ARGTYPE_STRING:
      found = 1;
      break;

    case TCL_OBJECT_ARGTYPE_BOOL:
      if (Tcl_GetBoolean (NULL, *argv, &itemp) == TCL_OK)
	found = 1;
      break;

    case TCL_OBJECT_ARGTYPE_DOUBLE:
      if (Tcl_GetDouble (NULL, *argv, &dtemp) == TCL_OK)
	found = 1;
      break;

    case TCL_OBJECT_ARGTYPE_INT:
    case TCL_OBJECT_ARGTYPE_UINT:
    case TCL_OBJECT_ARGTYPE_LONG:
    case TCL_OBJECT_ARGTYPE_ULONG:
      if (Tcl_GetInt (NULL, *argv, &itemp) == TCL_OK)
	found = 1;
      break;

    case TCL_OBJECT_ARGTYPE_PTR:
      if (TclObj_GetPointer (NULL, *argv, &vtemp) == TCL_OK)
	found = 1;
      break;

    default:
      /*
       * any other object
       */
      if (Tcl_GetCommandInfo (interp, *argv, &cmdinfo) == 1) {
	if (cmdinfo.clientData) {
	  /*
	   * check if object is compatible with given type
	   */
	  if (TclObj_Cast (interp, (Tcl_Object*) cmdinfo.clientData,
			   iter->ident) != NULL) {
	    found = 1;
	  }
	}
      }
      break;
    }
    if (found) break;
  }

  if (found) {
    return TclObj_SearchArgtree (interp, iter->down, scopeident,
				 argc-1, argv+1, result);
  }

  /*
   * argument didn't match
   */
      
  return NULL;
}

static const TclObj_Argtree *
TclObj_SearchScopedSignature (Tcl_Interp *interp,
			      const TclObj_Signature *here,
			      unsigned long scopeident,
			      const char *fname,
			      int argc, char *argv[],
			      int *result)
{
  int cres;

  if (here == NULL)
    return NULL;

  cres = _FP_stricmp ((char *) fname, (char *) here->name);

  if (cres < 0)
    return TclObj_SearchScopedSignature (interp, here->left,
					 scopeident, fname,
					 argc, argv, result);
  else if (cres > 0)
    return TclObj_SearchScopedSignature (interp, here->right,
					 scopeident, fname,
					 argc, argv, result);

  return TclObj_SearchArgtree (interp, here->argtree, scopeident, 
			       argc, argv, result);
}

/*
 * This function resolves a potential scope resolution operator
 */

const TclObj_Argtree *
TclObj_SearchSignature (Tcl_Interp *interp, const TclObj_Info *here,
			const char *fname, int argc, char *argv[],
			int *result)
{
  char tmpstr[256], *p2;
  const char *p1;
  unsigned long thescope=0;
  TclObj_Info *theinfo;

  p1 = fname;
  p2 = tmpstr;

  while (*p1 && *p1!=':') {
    *p2++ = *p1++;
  }
  *p2 = '\0';

  if (*p1++ == ':') {
    if (*p1 == ':') p1++;
    /*
     * we have a class name in tmpstr
     */
    if (tmpstr[0] == '\0') {
      Tcl_AppendResult (interp, "empty scope resolution", NULL);
      *result = TCL_ERROR;
      return NULL;
    }
    if ((theinfo = TclObj_GetClassInfo (interp, tmpstr)) == NULL) {
      Tcl_AppendResult (interp, "illegal scope resolution to class \"",
			tmpstr, "\": unknown type", NULL);
      *result = TCL_ERROR;
      return NULL;
    }

    if (!TclObj_IsCompatible (interp, here, theinfo->ident)) {
      Tcl_AppendResult (interp, "illegal scope resolution to class \"",
			tmpstr, "\" for object of type \"",
			here->name, "\"", NULL);
      *result = TCL_ERROR;
      return NULL;
    }

    thescope = theinfo->ident;
  }
  else if (p1 == fname) {
    Tcl_AppendResult (interp, "illegal function call: no member name given",
		      NULL);
    *result = TCL_ERROR;
    return NULL;
  }
  else {
    p1 = fname;
  }
  
  return TclObj_SearchScopedSignature (interp, here->members,
				       thescope, p1, argc, argv,
				       result);
}

/*
 * ----------------------------------------------------------------------
 * Dump a list of the available member functions.
 * ----------------------------------------------------------------------
 */

static const TclObj_Info *
TclObj_FindMemberOfScope (const TclObj_Argtree *here,
			  unsigned long scopeident)
{
  const TclObj_Info *result;

  if (here==NULL)
    return 0;
  
  if (here->ident==0 && 
      (scopeident == 0 || scopeident == here->provider->ident)) {
    return here->provider;
  }

  if ((result = TclObj_FindMemberOfScope (here->next, scopeident)) != NULL)
    return result;

  return TclObj_FindMemberOfScope (here->down, scopeident);
}

static int
TclObj_DumpMemberNames (Tcl_Interp *interp, const TclObj_Signature *here,
			unsigned long scopeident, const char *fname,
			int *matches)
{
  const TclObj_Info *provider;
  Tcl_DString blubber;

  if (here) {

    TclObj_DumpMemberNames (interp, here->left,  scopeident, fname, matches);

    if (fname==NULL || _FP_stricmp ((char*) fname, (char*) here->name)==0) {
      if (matches == NULL) {
	if (scopeident == 0) {
	  Tcl_AppendElement (interp, here->name);
	}
	else if ((provider =
		  TclObj_FindMemberOfScope (here->argtree,
					    scopeident)) != NULL) {
	  Tcl_DStringInit   (&blubber);
	  Tcl_DStringAppend (&blubber,
			     (char*) provider->name,
			     strlen (provider->name));
	  Tcl_DStringAppend (&blubber, "::", 2);
	  Tcl_DStringAppend (&blubber,
			     (char*) here->name,
			     strlen (here->name));
	  Tcl_AppendElement (interp, Tcl_DStringValue (&blubber));
	  Tcl_DStringFree   (&blubber);
	}
      }
      else {
	*matches += 1;
      }
    }

    TclObj_DumpMemberNames (interp, here->right, scopeident, fname, matches);
  }
  return TCL_OK;
}

int
TclObj_DumpMemberNames (Tcl_Interp *interp, const TclObj_Info *here,
			const char *fname, int *matches)
{
  unsigned long thescope=0;
  char tmpstr[256], *p2;
  TclObj_Info *theinfo;
  const char *p1;

  if (fname) {
    p1 = fname;
    p2 = tmpstr;

    while (*p1 && *p1!=':') {
      *p2++ = *p1++;
    }
    *p2 = '\0';

    if (*p1++ == ':') {
      if (*p1 == ':') p1++;
      /*
       * we have a class name in tmpstr
       */
      if (tmpstr[0] != '\0') {
	if ((theinfo = TclObj_GetClassInfo (interp, tmpstr)) == NULL) {
	  Tcl_AppendResult (interp, "illegal scope resolution to class \"",
			    tmpstr, "\": unknown type", NULL);
	  return TCL_ERROR;
	}
      
	if (!TclObj_IsCompatible (interp, here, theinfo->ident)) {
	  Tcl_AppendResult (interp, "illegal scope resolution to class \"",
			    tmpstr, "\" for object of type \"",
			    here->name, "\"", NULL);
	  return TCL_ERROR;
	}
	thescope = theinfo->ident;
      }
    }
    else {
      p1 = fname;
    }
  }

  if (matches) {
    *matches = 0;
  }

  if (fname) {
    p1 = (*p1=='\0' || *p1==':') ? (const char *) NULL : p1;
  }
  else {
    p1 = (const char *) NULL;
  }

  return TclObj_DumpMemberNames (interp, here->members,
				 thescope, p1, matches);
}

static int
TclObj_DumpArgtree (Tcl_Interp *interp, const TclObj_Argtree *here,
		    unsigned long scopeident, Tcl_DString *sofar,
		    int *matches, int *needcr)
{
  Tcl_DString mystring;
  TclObj_Info *theinfo;

  if (here==NULL)
    return TCL_OK;

  if (here->ident == 0 &&
      (scopeident == 0 || scopeident == here->provider->ident)) {
    if (matches == NULL) {
      if (*needcr) {
	Tcl_AppendResult  (interp, "\n", NULL);
      }
      Tcl_AppendResult  (interp, here->provider->name,
			 "::", Tcl_DStringValue (sofar),
			 (here->entry->desc)?" (":"",
			 (here->entry->desc)?here->entry->desc:"",
			 (here->entry->desc)?")":"",
			 NULL);
      *needcr = 1;
    }
    else {
      *matches += 1;
    }
  }

  if (here->next) {
    TclObj_DumpArgtree (interp, here->next, scopeident, sofar,
			matches, needcr);
  }

  if (here->down) {
    Tcl_DStringInit   (&mystring);
    Tcl_DStringAppend (&mystring,
		       Tcl_DStringValue  (sofar),
		       Tcl_DStringLength (sofar));

    switch (here->ident) {
    case TCL_OBJECT_ARGTYPE_VARARG:
      Tcl_DStringAppendElement (&mystring, "...");
      break;

    case TCL_OBJECT_ARGTYPE_STRING:
      Tcl_DStringAppendElement (&mystring, "string");
      break;

    case TCL_OBJECT_ARGTYPE_BOOL:
      Tcl_DStringAppendElement (&mystring, "bool");
      break;

    case TCL_OBJECT_ARGTYPE_DOUBLE:
      Tcl_DStringAppendElement (&mystring, "double");
      break;

    case TCL_OBJECT_ARGTYPE_INT:
      Tcl_DStringAppendElement (&mystring, "int");
      break;

    case TCL_OBJECT_ARGTYPE_UINT:
      Tcl_DStringAppendElement (&mystring, "uint");
      break;

    case TCL_OBJECT_ARGTYPE_LONG:
      Tcl_DStringAppendElement (&mystring, "long");
      break;

    case TCL_OBJECT_ARGTYPE_ULONG:
      Tcl_DStringAppendElement (&mystring, "ulong");
      break;

    case TCL_OBJECT_ARGTYPE_PTR:
      Tcl_DStringAppendElement (&mystring, "ptr");
      break;

    default: /* any other type */
      if ((theinfo = TclObj_GetClassInfo (interp, here->ident)) != NULL) {
	Tcl_DStringAppendElement (&mystring, (char*) theinfo->name);
      }
      else {
	Tcl_DStringAppendElement (&mystring, "oops?");
      }
      break;
    }
    TclObj_DumpArgtree (interp, here->down, scopeident, &mystring,
			matches, needcr);
    Tcl_DStringFree (&mystring);
  }
  return TCL_OK;
}

static int
TclObj_DumpScopedSignatures (Tcl_Interp *interp, const TclObj_Signature *here,
			     unsigned long scopeident, const char *fname,
			     int *matches, int *needcr)
{
  Tcl_DString sofar;

  if (here == NULL)
    return TCL_OK;

  TclObj_DumpScopedSignatures (interp, here->left,
			       scopeident, fname,
			       matches, needcr);

  if (fname==NULL || _FP_stricmp ((char*)fname, (char*)here->name) == 0) {
    Tcl_DStringInit   (&sofar);
    Tcl_DStringAppend (&sofar, (char*) here->name, strlen(here->name));
    TclObj_DumpArgtree(interp, here->argtree, scopeident, &sofar,
		       matches, needcr);
    Tcl_DStringFree   (&sofar);
  }

  TclObj_DumpScopedSignatures (interp, here->right,
			       scopeident, fname,
			       matches, needcr);

  return TCL_OK;
}

int
TclObj_DumpSignatures (Tcl_Interp *interp, const TclObj_Info *here,
		       const char *fname, int *matches)
{
  unsigned long thescope=0;
  char tmpstr[256], *p2;
  TclObj_Info *theinfo;
  const char *p1;
  int needcr=0;

  if (fname) {
    p1 = fname;
    p2 = tmpstr;

    while (*p1 && *p1!=':') {
      *p2++ = *p1++;
    }
    *p2 = '\0';

    if (*p1++ == ':') {
      if (*p1 == ':') p1++;
      /*
       * we have a class name in tmpstr
       */
      if (tmpstr[0] != '\0') {
	if ((theinfo = TclObj_GetClassInfo (interp, tmpstr)) == NULL) {
	  Tcl_AppendResult (interp, "illegal scope resolution to class \"",
			    tmpstr, "\": unknown type", NULL);
	  return TCL_ERROR;
	}
      
	if (!TclObj_IsCompatible (interp, here, theinfo->ident)) {
	  Tcl_AppendResult (interp, "illegal scope resolution to class \"",
			    tmpstr, "\" for object of type \"",
			    here->name, "\"", NULL);
	  return TCL_ERROR;
	}
	thescope = theinfo->ident;
      }
    }
    else {
      p1 = fname;
    }
  }

  if (matches) {
    *matches = 0;
  }

  if (fname) {
    p1 = (*p1=='\0' || *p1==':') ? (const char *) NULL : p1;
  }
  else {
    p1 = (const char *) NULL;
  }

  return TclObj_DumpScopedSignatures (interp, here->members,
				      thescope, p1, matches,
				      &needcr);
}
